import asyncio
import json

from .graphql import GraphQLQuery, GraphQLSubscription, GraphQLLiveQuery

from .mixins.basic_system_info import BasicSystemInfoMixin
from .mixins.scripts import ScriptsMixin
from .mixins.sequences import SequencesMixin
from .mixins.messages import MessagesMixin


class TritiumRemote(BasicSystemInfoMixin, ScriptsMixin, SequencesMixin, MessagesMixin):
    """
    A connecton to a remote Tritium system for interaction using its GraphQL API.

    Supports free-form GraphQL queries, mutations, subscriptions and live
    queries, or more specialised methods for controlling scripts, playing
    sequences etc
    """

    TIMEOUT_SECONDS = 10

    _connector = None

    def __init__(self, connector, on_message_handling_error=None):
        """
        Args:
            connector: Connection method (currently WebSocket only)
            on_message_handling_error (callback): For notification of errors during message handling, passed as `dict`.
        """
        self._connector = connector
        self._on_message_handling_error = on_message_handling_error
        self._next_request_id = 0
        self._graphql_operations_by_id = {}

    @property
    def connected(self):
        """
        True if connected to the Tritium system.
        """
        return self._connector and self._connector.connected

    async def connect(self):
        """
        Attempts to connect to the Tritium system
        """
        await self._connector.connect(self._on_message)

    async def disconnect(self):
        """
        Disconnects from the Tritium system (permanently).
        """
        await self._connector.disconnect()
        self._connector = None
        self._send = None

    ##########################################################################

    async def query(self, document, variables=None):
        """
        Sends a free-form GraphQL `query` or `mutation` request, and waits for the response.

        Args:
            document (str): A GraphQL `query` or `mutation` document
            variables (dict): Variable values by name, which must be declared in the document.

        Raises:
            `TimeoutError` if the Tritium system does not reply within `TIMEOUT_SECONDS`.

        Returns:
            The response from the system as a `dict` with `data` and possible `errors`.
        """
        request_id = await self._send_graphql_message(document, variables)

        query = GraphQLQuery(request_id)
        self._add_graphql_operation(request_id, query)

        # wait for response, or eventually raise asyncio.TimeoutError
        return await asyncio.wait_for(query.result, self.TIMEOUT_SECONDS)

    async def subscribe(self, document, variables=None):
        """
        Sends a GraphQL `subscription` request and provides a stream of results.

        Args:
            document (str): A GraphQL `subscription` document
            variables (dict): Variable values by name, which must be declared in the document.

        Returns:
            A `tritium_remote_py.graphql.subscription.GraphQLSubscription` object

        """
        request_id = await self._send_graphql_message(document, variables)

        async def cancel():
            self._remove_graphql_operation(request_id)
            await self._send_graphql_close_message(request_id)

        subscription = GraphQLSubscription(request_id, cancel)
        self._add_graphql_operation(request_id, subscription)

        return subscription

    async def live_query(self, document, variables=None, make_value=None):
        """
        Sends a GraphQL `subscription` request but returns an object for watching the
        response value.

        The GraphQL API is expected to provide a stream of updated values of a certain variable
        or structure.  By convention, live query subscriptions are named `live{Something}` to
        distinguish them from plain stream-of-data GraphQL subscriptions.

        Args:
            document (str): A GraphQL `subscription` document
            variables (dict): Variable values by name, which must be declared in the document.
            make_value (function): A function which creates the required value type from the raw response data

        Returns:
            A `tritium_remote_py.graphql.live_query.GraphQLLiveQuery` object

        """
        subscription = await self.subscribe(document, variables)
        live_query = GraphQLLiveQuery(subscription, make_value)
        # replaces subscription
        self._add_graphql_operation(subscription.request_id, live_query)
        return live_query

    ##########################################################################

    async def _send(self, msg):
        if self._connector:
            await self._connector.send(msg)

    def _on_message(self, msg):
        try:
            m = json.loads(msg)
        except Exception as e:
            self._on_message_error(error="failed to decode message JSON", exception=e)
            return

        try:
            message_type = m["type"]
            result = m["data"]
            request_id = m["request_id"]
        except KeyError as e:
            self._on_message_error(error="invalid message", exception=e)
            return

        if message_type == "graphql_response":
            self._on_graphql_response(request_id, result)
        else:
            self._on_message_error(error=f"unrecognised message type: {message_type}")

    def _on_message_error(self, **error):
        if self._on_message_handling_error:
            self.self._on_message_handling_error(error)

    async def _send_graphql_message(self, document, variables):
        request_id = self._next_request_id
        self._next_request_id += 1

        msg = json.dumps(
            {
                "type": "graphql",
                "request_id": request_id,
                "document": document,
                "variable_values": variables,
            }
        )

        await self._send(msg)

        return request_id

    async def _send_graphql_close_message(self, request_id):
        msg = json.dumps(
            {
                "type": "graphql_close",
                "request_id": request_id,
            }
        )

        await self._send(msg)

    def _add_graphql_operation(self, request_id, graphql_operation):
        self._graphql_operations_by_id[request_id] = graphql_operation

    def _remove_graphql_operation(self, request_id):
        try:
            del self._graphql_operations_by_id[request_id]
        except KeyError:
            pass

    def _on_graphql_response(self, request_id, result):
        try:
            op = self._graphql_operations_by_id[request_id]
            op.on_result(result)

            if not op.more_results_expected:
                self._remove_graphql_operation(request_id)
        except KeyError:
            pass



from enum import Enum

from ..exceptions import raise_if_graphql_errors


class ScriptAction(Enum):
    START = "START"
    STOP = "STOP"


class ScriptStatus(Enum):
    LOADED = "LOADED"
    RUNNING = "RUNNING"
    ERROR = "ERROR"


class ScriptsMixin:
    async def start_script(self, path):
        """
        Start executing the given script.

        *NB* The returned script status is the value after the `script` State
        Engine activity is started, but before that activity is picked up by
        the Scripts node to run the script.  Use `script_status` or
        `watch_script_status` to determine if the script actually runs OK.

        Args:
            path (str): The script path

        Returns:
            The script status (but see the note above).
        """
        return await self._trigger_script(path, ScriptAction.START)

    async def stop_script(self, path):
        """
        Stop the given script.

        Args:
            path (str): The script path

        Returns:
            The script status.
        """
        return await self._trigger_script(path, ScriptAction.STOP)

    async def _trigger_script(self, path, action):
        document = """
mutation ManuallyTriggerScript($input: ScriptTriggerInput!) {
    manuallyTriggerScript(input: $input) {
        script {
            status
        }
    }
}"""
        variables = {"input": {"path": path, "action": action.value}}
        response = await self.query(document, variables)

        raise_if_graphql_errors(response)

        status_text = response["data"]["manuallyTriggerScript"]["script"]["status"]

        return ScriptStatus[status_text]

    async def script_status(self, path):
        """
        Queries the given script's status.

        Args:
            path (str): The script path

        Returns:
            The script status.
        """
        document = """
query Script($path: String!) {
    script(path: $path) {
        status
    }
}"""
        variables = {"path": path}
        response = await self.query(document, variables)

        raise_if_graphql_errors(response)

        status_text = response["data"]["script"]["status"]

        return ScriptStatus[status_text]

    async def watch_script_status(self, path):
        """
        Watches the given script's status for changes.

        Args:
            path (str): The script path

        Returns:
            `tritium_remote_py.graphql.live_query.GraphQLLiveQuery`
        """

        document = """
subscription LiveScript($path: String!) {
    liveScript(path: $path) {
        status
    }
}"""
        variables = {"path": path}

        def make_value(response_data):
            return response_data["liveScript"]["status"]

        return await self.live_query(document, variables, make_value)



class TritiumRemoteError(Exception):
    """
    Exception thrown when something goes wrong using `tritium_remote_py`.
    """

    pass


class GraphQLErrors(TritiumRemoteError):
    """
    Raised when a GraphQL response contains errors.
    """

    def __init__(self, graphql_errors):
        super().__init__(f"GraphQL errors: {graphql_errors}")
        self._graphql_errors = graphql_errors


def raise_if_graphql_errors(graphql_response):
    """
    Raises `GraphQLErrors` if there are *any* errors in the response.

    *NB* This may not be the behaviour you want - in some circumstances it is
    perfectly reasonable for a GraphQL operation to return both data and errors.
    """
    try:
        errors = graphql_response["errors"]
        if errors:
            raise GraphQLErrors(errors)
    except KeyError:
        pass



from ..exceptions import raise_if_graphql_errors


class SequencesMixin:
    async def play_sequence(self, project_path):
        """
        Play the given sequence.

        Args:
            project_path (str): The Tritium Animator project asset path
        """
        document = """
mutation PlaySequence($input: PlaySequenceInput!) {
    playSequence(input: $input) {
        id
    }
}
"""
        variables = {"input": {"projectPath": project_path}}
        response = await self.query(document, variables)

        raise_if_graphql_errors(response)

        # TODO what is this ID?
        return response["data"]["playSequence"]["id"]



import asyncio
import json


class MockConnector:
    _connected = False

    def __init__(self, auth_token=None):
        self._auth_token = auth_token
        self._mock_replies = asyncio.Queue()
        self._mock_reply_tasks = []

    @property
    def connected(self):
        return self._connected

    async def connect(self, on_message):
        self._on_message = on_message
        self._connected = True

    async def disconnect(self):
        self._connected = False

    async def send(self, msg):
        self.last_sent_msg = json.loads(msg)

        self.send_mock_replies()

    def add_mock_reply(self, **msg):
        self._mock_replies.put_nowait(msg)

    def send_mock_replies(self):
        while True:
            try:
                msg = self._mock_replies.get_nowait()
                task = asyncio.create_task(self.send_mock_reply(**msg))
                self._mock_reply_tasks.append(task)
            except asyncio.QueueEmpty:
                break

    async def send_mock_reply(self, **msg):
        self._on_message(json.dumps(msg))



import asyncio

from .operation import GraphQLOperation


class GraphQLLiveQuery(GraphQLOperation):
    """
    A GraphQL live query, a non-standard addition to the GraphQL world.
    Basically a query that receives updates to the queried value.
    """

    _last_value = None

    def __init__(self, subscription, make_value):
        super().__init__(subscription.request_id)
        self._subscription = subscription
        if make_value:
            self._make_value = make_value
        else:
            self._make_value = lambda d: d
        self._value_changed_event = asyncio.Event()

    async def cancel(self):
        """
        Cancel the live query.
        """
        await self._subscription.cancel()

    def on_result(self, result):
        """
        @private
        """
        self._subscription.on_result(result)
        self._last_value = self._make_value(result["data"])
        self._value_changed_event.set()

    @property
    def more_results_expected(self):
        """
        @private
        """
        return self._subscription.more_results_expected

    @property
    async def value_changed(self):
        """
        Awaitable event indicating that the value has been updated.
        """
        await self._value_changed_event.wait()
        self._value_changed_event.clear()

    @property
    def value(self):
        """
        The current query value.
        """
        return self._last_value



"""
A pure Python library for interacting with a Tritium system.

Example - querying basic system info
```python
import os
import asyncio
import tritium_remote_py


async def main():
    auth_token = os.environ["TRITIUM_AUTH_TOKEN"]
    host = os.environ.get("TRITIUM_HOST", "localhost")

    tritium = await tritium_remote_py.connect_websocket(
        url=f"ws://{host}:1234",
        auth_token=auth_token,
        description="Python tritium-remote example - basic system info",
    )

    info = await tritium.basic_system_info()
    print("Serial:", info.serial)
    print("Version:", info.version)
    print("Name:", info.name)

    await tritium.disconnect()


asyncio.run(main())

```

See `tritium_remote_py.connect.connect` to get started.
"""

from .client import TritiumRemote  # noqa: F401
from .connect import connect_websocket  # noqa: F401



from .client import TritiumRemote
from .connection import WebSocketConnector
from .exceptions import TritiumRemoteError


async def connect(**kwargs):
    """
    Connects to a remote Tritium system, using the supplied keyword arguments
    to choose the connection method.

    Currently only WebSocket connections are supported. See `connect_websocket`
    below for more details.

    Returns:
        The connected `tritium_remote_py.client.TritiumRemote`.
    """

    try:
        url = kwargs["url"]
        auth_token = kwargs["auth_token"]
    except KeyError as e:
        raise TritiumRemoteError("Missing connection argument") from e

    description = kwargs.get("description")
    return await connect_websocket(url, auth_token, description)


async def connect_websocket(url, auth_token, description=None):
    """
    Connects to a remote Tritium system via WebSocket.

    Args:
        url (str): The Tritium system Gateway URL.
        auth_token (str): A JWT granting access to the given system.
        description (str): Human readable description of the connection.

    Returns:
        The connected `tritium_remote_py.client.TritiumRemote`.

    """

    connector = WebSocketConnector(url, auth_token, description)
    tritium_remote = TritiumRemote(connector)

    await tritium_remote.connect()

    return tritium_remote



class GraphQLOperation:
    """
    GraphQL operation base class
    """

    def __init__(self, request_id):
        self._request_id = request_id

    @property
    def request_id(self):
        """
        The operation request ID.
        """
        return self._request_id

    def on_result(self, result):
        """
        @private
        """
        # must be overidden
        raise NotImplementedError()

    @property
    def more_results_expected(self):
        """
        @private
        """
        # must be overidden
        raise NotImplementedError()



import asyncio
import websockets
import json


class WebSocketConnector:
    _ws = None
    _listen_task = None

    def __init__(self, url, auth_token, description):
        self._url = url
        self._auth_token = auth_token
        self._description = description

    @property
    def connected(self):
        return self._ws and self._ws.open

    @property
    def _headers(self):
        auth_token = self._auth_token
        metadata = json.dumps(
            {"session_type": "graphql", "description": self._description}
        )
        return {"x-tritium-token": auth_token, "x-tritium-session-metadata": metadata}

    async def connect(self, on_message):
        ws = await websockets.connect(self._url, extra_headers=self._headers)
        self._ws = ws

        async def listen():
            while True:
                msg = await ws.recv()
                on_message(msg)

        self._listen_task = asyncio.create_task(listen())

    async def disconnect(self):
        ws = self._ws
        if ws:
            self._ws = None
            await ws.close()

        if self._listen_task:
            self._listen_task.cancel()
            self._listen_task = None

    async def send(self, msg):
        if self._ws:
            await self._ws.send(msg)



import asyncio

from .operation import GraphQLOperation


class GraphQLQuery(GraphQLOperation):
    """
    A simple GrapQL query for data.
    """

    def __init__(self, request_id):
        super().__init__(request_id)

        loop = asyncio.get_running_loop()
        self._future = loop.create_future()

    def on_result(self, result):
        """
        @private
        """
        self._future.set_result(result)

    @property
    def more_results_expected(self):
        """
        @private
        """
        return False

    @property
    async def result(self):
        """
        The query result as a `dict` with `data` and possibly `errors`.
        """
        await self._future
        return self._future.result()



import os
from setuptools import setup, find_packages


def read(file_name):
    return open(os.path.join(os.path.dirname(__file__), file_name)).read()


setup(
    name="tritium-remote-py",
    version=read("VERSION").strip(),
    description="Tritium remote interaction.",
    long_description=read("README.rst"),
    license="GPL",
    keywords="tritium robotics",
    url="http://wiki.engineeredarts.co.uk/tritium",
    author="Engineered Arts Ltd",
    author_email="tritium@engineeredarts.co.uk",
    install_requires=["websockets>=12.0"],
    packages=find_packages(exclude=["tests"]),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Environment :: No Input/Output (Daemon)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering",
        "Framework :: Robot Framework :: Library",
    ],
    setup_requires=["setuptools"],
)



from collections import namedtuple

from ..exceptions import raise_if_graphql_errors


BasicSystemInfo = namedtuple("BasicSystemInfo", ["serial", "version", "name"])


class BasicSystemInfoMixin:
    async def basic_system_info(self):
        """
        Queries the system for basic information:
        - Serial
        - Tritium version
        - Name
        """

        response = await self.query(
            """
query BasicSystemInfo {
    system {
        serial
        version
        name
    }
}            
        """
        )

        raise_if_graphql_errors(response)

        info = response["data"]["system"]
        return BasicSystemInfo(**info)



from ..exceptions import raise_if_graphql_errors


class MessagesMixin:
    async def post_message(self, channel, message):
        """
        Post a message to a named channel for a script to pick up.

        Args:
            channel (str): The channel name
            message (any JSON encodable value): The message
        """
        document = """
mutation PostMessage($input: PostMessageInput!) {
   postMessage(input: $input) 
}
"""
        variables = {"input": {"channel": channel, "message": message}}

        response = await self.query(document, variables)

        raise_if_graphql_errors(response)



import asyncio

from .operation import GraphQLOperation


class GraphQLSubscription(GraphQLOperation):
    """
    A GraphQL subscription providing a stream of results.
    """

    def __init__(self, request_id, cancel):
        super().__init__(request_id)
        self._cancel = cancel
        self._queue = asyncio.Queue()

    @property
    async def results(self):
        """
        Asynchronous iterator of results.

        Each is a `dict` with `data` and possible `errors`.
        """

        while True:
            result = await self._queue.get()
            yield result

    async def cancel(self):
        """
        Cancel the subscription.
        """
        await self._cancel()

    def on_result(self, result):
        """
        @private
        """
        self._queue.put_nowait(result)

    @property
    def more_results_expected(self):
        """
        @private
        """
        return True
